#ifndef SRC_TIME_STEPPING_ADAPTIVETIMESTEPPER_HH
#define SRC_TIME_STEPPING_ADAPTIVETIMESTEPPER_HH

#include <fstream>

//#include "../multi-threading/task.hh"
#include "coupledtimestepper.hh"

struct IterationRegister {
  void registerCount(FixedPointIterationCounter count);
  void registerFinalCount(FixedPointIterationCounter count);

  void reset();

  FixedPointIterationCounter totalCount;
  FixedPointIterationCounter finalCount;
};

template <class Factory, class ContactNetwork, class Updaters, class ErrorNorms>
class AdaptiveTimeStepper {
  struct UpdatersWithCount {
    Updaters updaters;
    FixedPointIterationCounter count;
  };

  using Vector = typename Factory::Vector;
  using Matrix = typename Factory::Matrix;
  using IgnoreVector = typename Factory::BitVector;
  //using ConvexProblem = typename Factory::ConvexProblem;
  //using Nonlinearity = typename Factory::Nonlinearity;

  using NBodyAssembler = typename ContactNetwork::NBodyAssembler;

  using MyCoupledTimeStepper = CoupledTimeStepper<Factory, NBodyAssembler, Updaters, ErrorNorms>;

  using GlobalFriction = typename MyCoupledTimeStepper::GlobalFriction;
  using BitVector = typename MyCoupledTimeStepper::BitVector;
  using ExternalForces = typename MyCoupledTimeStepper::ExternalForces;

public:
  AdaptiveTimeStepper(
                      Dune::ParameterTree const &parset,
                      ContactNetwork& contactNetwork,
                      const IgnoreVector& ignoreNodes,
                      GlobalFriction& globalFriction,
                      const std::vector<const BitVector*>& bodywiseNonmortarBoundaries,
                      Updaters &current,
                      double relativeTime,
                      double relativeTau,
                      ExternalForces& externalForces,
                      const ErrorNorms& errorNorms,
                      std::function<bool(Updaters &, Updaters &)> mustRefine);

  bool reachedEnd();

  auto makeLinearSolver() const;

  IterationRegister advance();

  double relativeTime_;
  double relativeTau_;

private:
  void setDeformation(const Updaters& updaters);

  NBodyAssembler step(const NBodyAssembler& oldNBodyAssembler) const;


  template <class LinearSolver>
  UpdatersWithCount step(const Updaters& oldUpdaters, const NBodyAssembler& nBodyAssembler,
                         std::shared_ptr<LinearSolver>& linearSolver,
                         double rTime, double rTau);

  double finalTime_;
  Dune::ParameterTree const &parset_;
  ContactNetwork& contactNetwork_;
  const IgnoreVector& ignoreNodes_;

  GlobalFriction& globalFriction_;
  const std::vector<const BitVector*>& bodywiseNonmortarBoundaries_;

  Updaters &current_;
  UpdatersWithCount R1_;
  ExternalForces& externalForces_;
  std::function<bool(Updaters &, Updaters &)> mustRefine_;
  const ErrorNorms& errorNorms_;

  IterationRegister iterationRegister_;
};

#endif
